@@ -45,6 +45,7 @@ import java.util.Objects;
 import java.util.function.Predicate;
 
 import jdk.vm.ci.code.Architecture;
+import jdk.vm.ci.code.CompilationRequest;
 import jdk.vm.ci.code.CompilationRequestResult;
 import jdk.vm.ci.code.CompiledCode;
 import jdk.vm.ci.code.InstalledCode;
@@ -366,8 +367,10 @@ public final class HotSpotJVMCIRuntime implements JVMCIRuntime {
         /**
          * Parses all system properties starting with {@value #JVMCI_OPTION_PROPERTY_PREFIX} and
          * initializes the options based on their values.
+         *
+         * @param compilerToVm
          */
-        static void parse() {
+        static void parse(CompilerToVM compilerToVm) {
             Map<String, String> savedProps = jdk.vm.ci.services.Services.getSavedProperties();
             for (Map.Entry<String, String> e : savedProps.entrySet()) {
                 String name = e.getKey();
@@ -382,14 +385,17 @@ public final class HotSpotJVMCIRuntime implements JVMCIRuntime {
                             }
                         }
                         Formatter msg = new Formatter();
-                        msg.format("Could not find option %s", name);
+                        msg.format("Error parsing JVMCI options: Could not find option %s", name);
                         if (!matches.isEmpty()) {
                             msg.format("%nDid you mean one of the following?");
                             for (String match : matches) {
                                 msg.format("%n    %s=<value>", match);
                             }
                         }
-                        throw new IllegalArgumentException(msg.toString());
+                        msg.format("%nError: A fatal exception has occurred. Program will exit.%n");
+                        byte[] msgBytes = msg.toString().getBytes();
+                        compilerToVm.writeDebugOutput(msgBytes, 0, msgBytes.length, true, true);
+                        compilerToVm.callSystemExit(1);
                     } else if (value instanceof Option) {
                         Option option = (Option) value;
                         option.init(e.getValue());
@@ -495,7 +501,7 @@ public final class HotSpotJVMCIRuntime implements JVMCIRuntime {
         }
 
         // Initialize the Option values.
-        Option.parse();
+        Option.parse(compilerToVm);
 
         String hostArchitecture = config.getHostArchitectureName();
 
@@ -666,6 +672,19 @@ public final class HotSpotJVMCIRuntime implements JVMCIRuntime {
         return null;
     }
 
+    static class ErrorCreatingCompiler implements JVMCICompiler {
+        private final RuntimeException t;
+
+        ErrorCreatingCompiler(RuntimeException t) {
+            this.t = t;
+        }
+
+        @Override
+        public CompilationRequestResult compileMethod(CompilationRequest request) {
+            throw t;
+        }
+    }
+
     @Override
     public JVMCICompiler getCompiler() {
         if (compiler == null) {
@@ -673,11 +692,19 @@ public final class HotSpotJVMCIRuntime implements JVMCIRuntime {
                 if (compiler == null) {
                     assert !creatingCompiler : "recursive compiler creation";
                     creatingCompiler = true;
-                    compiler = compilerFactory.createCompiler(this);
-                    creatingCompiler = false;
+                    try {
+                        compiler = compilerFactory.createCompiler(this);
+                    } catch (RuntimeException t) {
+                        compiler = new ErrorCreatingCompiler(t);
+                    } finally {
+                        creatingCompiler = false;
+                    }
                 }
             }
         }
+        if (compiler instanceof ErrorCreatingCompiler) {
+            throw ((ErrorCreatingCompiler) compiler).t;
+        }
         return compiler;
     }
 
@@ -1128,7 +1155,8 @@ public final class HotSpotJVMCIRuntime implements JVMCIRuntime {
      *             the length of the array returned by {@link #registerNativeMethods}
      */
     public boolean attachCurrentThread(boolean asDaemon) {
-        return compilerToVm.attachCurrentThread(asDaemon);
+        byte[] name = Services.IS_IN_NATIVE_IMAGE ? Thread.currentThread().getName().getBytes() : null;
+        return compilerToVm.attachCurrentThread(name, asDaemon);
     }
 
     /**
